# coding=utf-8
# Copyright (c) Andreas Panteli, Jonas Teuwen
"""Loss module for the Exclusive Cross Entropy."""
import math
from typing import Any, Callable, Optional, Union, Tuple, Type

import torch
from torch import Tensor
from torch.nn import BCELoss, BCEWithLogitsLoss, CrossEntropyLoss


def get_sigmoid_annealing_function(epoch_checkpoint: int, temperature: float) -> Callable:
    """
    Implements sigmoid annealing given a checkpoint (for the turning point - middle) and temperature (speed of change)
    """

    def _sigmoid(data):
        """Ordinary sigmoid function"""
        return 1 / (1 + math.exp(-data))

    def _annealing(epoch: int, threshold: float) -> float:
        """Annealing centred around epoch_checkpoint with a decline rate depending on temperature"""
        return _sigmoid((epoch - epoch_checkpoint) / temperature) * threshold

    return _annealing


def get_focal_weighing_function(focal_loss_alpha: float, focal_loss_gamma: float) -> Callable:
    """
    Implements a focal reweighing for loss based on the prediction confidence.
    Assumes loss is pass is defined at it's output as -log(prediction)
    """

    def _focal_weighing(loss: Tensor) -> Tensor:
        loss_p = torch.exp(-loss)
        return focal_loss_alpha * ((1.0 - loss_p) ** focal_loss_gamma) * loss

    return _focal_weighing


class ExclusiveLoss:  # pylint: disable=R0902
    """
    Implements an abstract class of the exclusivity cross entropy as described in the Sparse-shot Learning with
    Exclusive Cross-Entropy for Extremely Many Localisations paper[1].

    Default values are defined as in the original work.

    The different versions, namely BECELoss, BECEWithLogitsLoss, ExclusiveCrossEntropyLoss, implement a similar
    loss as their original pytorch non-exlusive versions for easy/familiar use cases.

    Reference:
        [1] Panteli, A., Teuwen, J., Horlings, H. and Gavves, E.; Sparse-shot Learning with Exclusive Cross-Entropy for
        Extremely Many Localisations; Proceedings of the IEEE/CVF International Conference on Computer Vision (ICCV),
        2021, pp. 2813-2823
    """

    def __init__(  # pylint: disable=R0913
        self,
        baseloss: Union[Type[BCELoss], Type[BCEWithLogitsLoss], Type[CrossEntropyLoss]],
        target_dtype: torch.dtype,
        confidence_estimator: Callable,
        exclusivity_threshold: float = 0.5,
        background_sampling_threshold: float = 0.5,
        exclusivity_threshold_annealing: Callable = get_sigmoid_annealing_function(50, 10),
        background_sampling_annealing: Callable = get_sigmoid_annealing_function(150, 10),
        focal_loss_parameters: Tuple[float, float] = (0.2, 0.1),
        epoch: int = 0,
    ) -> None:
        # super(ExclusiveLoss, self).__init__()
        self.baseloss = baseloss
        self.target_dtype = target_dtype
        self.confidence_estimator = confidence_estimator
        self.exclusivity_threshold = exclusivity_threshold
        self.background_sampling_threshold = background_sampling_threshold
        self.exclusivity_threshold_annealing = exclusivity_threshold_annealing
        self.background_sampling_annealing = background_sampling_annealing
        self.focal_loss_alpha, self.focal_loss_gamma = focal_loss_parameters
        self.epoch = epoch
        self.unannotated_mapping = self.all_labels_as_unannotated

        self.focal_loss = get_focal_weighing_function(self.focal_loss_alpha, self.focal_loss_gamma)

    def set_epoch(self, epoch: int) -> None:
        """Set the epoch at the current state of training for use in the annealing functions"""
        self.epoch = epoch

    @staticmethod
    def get_binary_confidence(predictions: Tensor) -> Tensor:
        """
        For binary predictions, i.e.: object vs. background, the object confidence output is equivalent
        to the prediction confidence. Hence, no change is made.

        Example:
        In a detection task, there will be n bounding boxes with likely properties x,y,w,h,class and confidence, where
        confidence indicates the perceived probability there exists an object/box in this area.
        """
        return predictions

    @staticmethod
    def get_multiclass_confidence(predictions: Tensor) -> Tensor:
        """
        For multiple output class predictions, the confidence is defined as the difference of the top two classes. As
        originally proposed by [1].
        """
        top_2 = torch.topk(predictions, k=2, dim=1)[0]
        predictions = torch.abs(top_2[:, 0, ...] - top_2[:, 1, ...])
        return predictions

    @staticmethod
    def zeroth_label_as_unannotated(targets: Tensor) -> Tensor:
        r"""
        Input of targets must have shape as follows
        :math:`(N)` where each value is :math:`0 \leq \text{targets}[i] \leq C-1`, or
        :math:`(N, d_1, d_2, ..., d_K)` with :math:`K \geq 1` in the case of K-dimensional loss.

        :param targets: matrix containing the class labels as integers
        :return: a mask indicating where the unannotated samples are
        """
        return targets == 0

    @staticmethod
    def all_labels_as_unannotated(targets: Tensor) -> Tensor:
        r"""
        Input of targets must have shape as follows
        :math:`(N)` where each value is :math:`0 \leq \text{targets}[i] \leq C-1`, or
        :param targets: matrix containing the class labels as integers
        :return: a mask indicating where the unannotated samples are
        """
        return torch.ones(len(targets))

    def set_unannotated_mapping(self, mapping: Callable) -> None:
        """
        Sets the mapping function, given targets, for the unlabelled class
        :param mapping: function to be used when identifying which samples belong to the unannotated cases
        """
        self.unannotated_mapping = mapping

    @staticmethod
    def sample_from_background(predictions: Tensor) -> Tensor:
        """
        Implements sampling across the prediction samples.
        Returns vector values in the range [0, 1] drawn from a uniform distribution.
        """
        return predictions.float().new(predictions.shape).uniform_()

    def forward_(
        self, input: Tensor, target: Tensor, epoch: int, unannotated_mapping: Callable  # pylint: disable=W0622
    ) -> Tensor:

        """
        Uses the exclusivity threshold and annealing functions as described in the Sparse-shot Learning with Exclusive
        Cross-Entropy for Extremely Many Localisations paper[1] to produce the exclusive cross entropy loss.

        :param unannotated_mapping: mapping function to retrieve location of unannotated areas/samples. E.g.: for a
            binary mask this can be a lambda x: x == 0.
        :param input:  representing the unlabelled samples, a list of bounding box confidences of shape BxN, or a list
            of pixels from re-shaped images of shape Bx(HxW), or classification predictions of shape B or multiclass
            input of shape BxC..., where C is the number of classes.
        :param target: label masks of shape BxHxW, or a list of bounding box target confidence of shape BxN, or
            classification targets of shape B.
        :param epoch: the epoch integer value of the training phase
        :return: exclusive cross-entropy
        """
        cross_entropy = self.baseloss_forward(input, target)

        # Background (uniform) sampling. Assuming all input is from the background class
        unannotated_mask = unannotated_mapping(target).bool().to(target.device)
        background_sampling = (
            self.sample_from_background(unannotated_mask)
            <= self.background_sampling_annealing(epoch, self.background_sampling_threshold)
        ) * unannotated_mask

        # Exclusivity threshold
        confidence = self.confidence_estimator(input)
        exclusivity_mask = confidence <= self.exclusivity_threshold_annealing(epoch, self.exclusivity_threshold)
        background_sampling *= exclusivity_mask

        # Focal re-weighing and sampling
        unannotated_loss = self.focal_loss(cross_entropy) * background_sampling
        annotated_loss = cross_entropy * (~unannotated_mask * 1.0)
        ece_loss = unannotated_loss + annotated_loss
        return ece_loss

    def forward(
        self,
        input: Tensor,  # pylint: disable=W0622
        target: Optional[Tensor] = None,
        epoch: Optional[int] = None,
        unannotated_mapping: Optional[Callable] = None,
    ) -> Tensor:
        """
        Overwrites the original forward function of the cross entropy loss (in all its variants: CrossEntropyLoss,
        BCELoss, BCEWithLogitsLoss) in order to be used in the familiar way.

        Assumes default zero tensors for the target and assigns the uses the most updated epoch value (either given
        as input or using the pre-set value)

        :param unannotated_mapping: mapping function to retrieve location of unannotated areas/samples. E.g.: for a
            binary mask this can be a lambda x: x == 0.
        :param epoch: the current epoch number of training used for annealing parameters of the ECE loss
        :param input: representing the unlabelled samples, a list of bounding box confidences of shape BxN, or a list
            of pixels from re-shaped images of shape Bx(HxW), or classification predictions of shape B or multiclass
            input of shape BxC..., where C is the number of classes.
        :param target: Optional parameter for the target ground truth values in the same shape as the expected input
            of the (binary) cross entropy functions. If not given, the assumed background label is zero.
        :return: the exclusive cross entropy loss value
        """
        if target is None:
            target = input.new_zeros(len(input), dtype=self.target_dtype)

        if epoch is None:
            epoch = self.epoch

        if unannotated_mapping is None:
            unannotated_mapping = self.unannotated_mapping

        return self.forward_(input, target, epoch, unannotated_mapping)

    def baseloss_forward(self, input: Tensor, target: Tensor) -> Tensor:  # pylint: disable=W0622
        """Implements the original CE variant forward function"""
        raise NotImplementedError


class BECELoss(ExclusiveLoss, BCELoss):
    """
    Binary Exclusive cross-entropy without activation loss as described in the Sparse-shot Learning with
    Exclusive Cross-Entropy for Extremely Many Localisations paper[1].
    It is a subclass of the abstract class ExclusiveLoss.

    Reference:
        [1] Panteli, A., Teuwen, J., Horlings, H. and Gavves, E., 2021. Sparse-shot Learning with Exclusive
        Cross-Entropy for Extremely Many Localisations. arXiv preprint arXiv:2104.10425.

    Shape:
        - Input: :math:`(N, *)` where :math:`*` means, any number of additional
          dimensions
        - Target: :math:`(N, *)`, same shape as the input
        - Output: scalar. If :attr:`reduction` is ``'none'``, then :math:`(N, *)`, same
          shape as input.

    Example:
        >>> m = nn.Sigmoid()
        >>> loss = nn.BECELoss()
        >>> input = torch.randn(3, requires_grad=True)
        >>> target = torch.empty(3).random_(2)
        >>> output = loss(m(input), target)
        >>> output.mean().backward()
    """

    def __init__(  # pylint: disable=R0913
        self,
        weight: Optional[Tensor] = None,
        size_average: Any = None,
        reduce: Any = None,
        reduction: str = "none",
        **kwargs,
    ) -> None:  # pylint: disable=R0801
        BCELoss.__init__(self, weight=weight, size_average=size_average, reduce=reduce, reduction=reduction)
        ExclusiveLoss.__init__(
            self, baseloss=BCELoss, target_dtype=torch.float, confidence_estimator=self.get_binary_confidence, **kwargs
        )

    def baseloss_forward(self, input: Tensor, target: Tensor) -> Tensor:  # pylint: disable=W0622
        return BCELoss.forward(self, input, target)


class BECEWithLogitsLoss(ExclusiveLoss, BCEWithLogitsLoss):
    """
    Binary Exclusive cross-entropy with sigmoid activation loss as described in the Sparse-shot Learning with
    Exclusive Cross-Entropy for Extremely Many Localisations paper[1].
    It is a subclass of the abstract class ExclusiveLoss.

    Reference:
        [1] Panteli, A., Teuwen, J., Horlings, H. and Gavves, E., 2021. Sparse-shot Learning with Exclusive
        Cross-Entropy for Extremely Many Localisations. arXiv preprint arXiv:2104.10425.

    Shape:
        - Input: :math:`(N, *)` where :math:`*` means, any number of additional dimensions
        - Target: :math:`(N, *)`, same shape as the input
        - Output: scalar. If :attr:`reduction` is ``'none'``, then :math:`(N, *)`, same
          shape as input.

    Example:
        >>> loss = BECEWithLogitsLoss()
        >>> batch_size = torch.randint(32, (1, 1)).squeeze()
        >>> input = torch.rand(batch_size, requires_grad=True)
        >>> output = loss(input).mean() # Assumes target as a default zero tensor
        >>> output.backward()
    """

    def __init__(  # pylint: disable=R0913
        self,
        weight: Optional[Tensor] = None,
        size_average: Any = None,
        reduce: Any = None,
        reduction: str = "none",
        pos_weight: Optional[Tensor] = None,
        **kwargs,
    ):  # pylint: disable=R0801
        BCEWithLogitsLoss.__init__(
            self, weight=weight, size_average=size_average, reduce=reduce, reduction=reduction, pos_weight=pos_weight
        )
        ExclusiveLoss.__init__(
            self,
            baseloss=BCEWithLogitsLoss,
            target_dtype=torch.float,
            confidence_estimator=self.get_binary_confidence,
            **kwargs,
        )

    def baseloss_forward(self, input: Tensor, target: Tensor) -> Tensor:  # pylint: disable=W0622
        return BCEWithLogitsLoss.forward(self, input, target)


class ExclusiveCrossEntropyLoss(ExclusiveLoss, CrossEntropyLoss):
    r"""
    Exclusive cross-entropy loss as described in the Sparse-shot Learning with Exclusive Cross-Entropy for
    Extremely Many Localisations paper[1]. It is a subclass of the abstract class ExclusiveLoss.

    Reference:
        [1] Panteli, A., Teuwen, J., Horlings, H. and Gavves, E., 2021. Sparse-shot Learning with Exclusive
        Cross-Entropy for Extremely Many Localisations. arXiv preprint arXiv:2104.10425.

    Shape:
        - Input: :math:`(N, C)` where `C = number of classes`, or
          :math:`(N, C, d_1, d_2, ..., d_K)` with :math:`K \geq 1`
          in the case of `K`-dimensional loss.
        - Target: :math:`(N)` where each value is :math:`0 \leq \text{targets}[i] \leq C-1`, or
          :math:`(N, d_1, d_2, ..., d_K)` with :math:`K \geq 1` in the case of
          K-dimensional loss.
        - Output: scalar.
          If :attr:`reduction` is ``'none'``, then the same size as the target:
          :math:`(N)`, or
          :math:`(N, d_1, d_2, ..., d_K)` with :math:`K \geq 1` in the case
          of K-dimensional loss.

    Example:
        >>> loss = ExclusiveCrossEntropyLoss()
        >>> epoch = torch.randint(300, (1, 1)).squeeze()
        >>> batch_size = torch.randint(32, (1, 1)).squeeze()
        >>> num_classes = torch.randint(100, (1, 1)).squeeze()
        >>> input = torch.randn(batch_size, num_classes, requires_grad=True)
        >>> target = torch.empty(batch_size, dtype=torch.long).random_(num_classes)
        >>> output = loss(input, target, epoch)
        >>> output = torch.mean(output)  # If reduction is left to the default of 'none'
        >>> output.backward()
    """

    def __init__(  # pylint: disable=R0913
        self,
        weight: Optional[Tensor] = None,
        size_average: Any = None,
        ignore_index: int = -100,
        reduce: Any = None,
        reduction: str = "none",
        **kwargs,
    ) -> None:  # pylint: disable=R0801
        CrossEntropyLoss.__init__(
            self,
            weight=weight,
            size_average=size_average,
            ignore_index=ignore_index,
            reduce=reduce,
            reduction=reduction,
        )
        ExclusiveLoss.__init__(
            self,
            baseloss=CrossEntropyLoss,
            target_dtype=torch.long,
            confidence_estimator=self.get_multiclass_confidence,
            **kwargs,
        )
        self.set_unannotated_mapping(self.zeroth_label_as_unannotated)

    def baseloss_forward(self, input: Tensor, target: Tensor) -> Tensor:  # pylint: disable=W0622
        return CrossEntropyLoss.forward(self, input, target)



# coding=utf-8
# Copyright (c) Andreas Panteli, Jonas Teuwen

"""Test the loss function facility classes."""

import torch
from torch.nn.modules.loss import BCELoss, BCEWithLogitsLoss, CrossEntropyLoss

from ece_loss.loss import BECELoss, BECEWithLogitsLoss, ExclusiveCrossEntropyLoss


def _identity_mapping_single_input(data):
    """Used for disabling components of ECE loss (during debugging/testing)"""
    return data


def _identity_mapping_double_input(_, label):
    """Used for disabling components of ECE loss (during debugging/testing)"""
    return label


def _identity_mapping_single_input_torch_ones(data):
    """Used for disabling components of ECE loss (during debugging/testing)"""
    return torch.ones(len(data))


def _identity_mapping_single_input_torch_zeros(data):
    """Used for disabling components of ECE loss (during debugging/testing)"""
    return torch.zeros(len(data))


def test_ce_output():  # pylint: disable=R0914
    """Test the outpout of the ECE if the exclusivity condition is diables, i.e. == CE"""
    ece_losses = [BECELoss, BECEWithLogitsLoss, ExclusiveCrossEntropyLoss]
    ce_losses = [BCELoss, BCEWithLogitsLoss, CrossEntropyLoss]

    linear_input_0_1 = torch.Tensor(range(11)) / 10
    linear_input_minus1_1 = (torch.Tensor(range(11)) - 5) / 5
    linear_multiclass_input = torch.cat((linear_input_0_1.unsqueeze(1), linear_input_minus1_1.unsqueeze(1)), dim=1)
    inputs = [linear_input_0_1, linear_input_minus1_1, linear_multiclass_input]
    targets = [
        torch.zeros(len(inputs[0])).float(),
        torch.zeros(len(inputs[1])).float(),
        torch.zeros(len(inputs[2])).long(),
    ]

    identity_mapping = _identity_mapping_double_input
    for ece_loss, ce_loss, input_sample, target in zip(ece_losses, ce_losses, inputs, targets):
        ce_loss_from_ece = ece_loss(
            exclusivity_threshold_annealing=identity_mapping,
            exclusivity_threshold=1,
            background_sampling_annealing=identity_mapping,
            background_sampling_threshold=1,
            focal_loss_parameters=(1, 0),
        )
        ce_loss_from_ece.sample_from_background = _identity_mapping_single_input_torch_ones
        ce_loss_no_reduction = ce_loss(reduction="none")

        ce_out = ce_loss_no_reduction(input_sample, target)
        ece_out = ce_loss_from_ece(input_sample, target)
        if not torch.allclose(ce_out, ece_out):
            print(ce_out, ece_out)
        assert torch.allclose(ce_out, ece_out)


def test_input_output():
    """Given specific input the output should be as pre-defined here"""
    linear_input_0_1 = torch.Tensor(range(11)) / 10
    linear_input_minus1_1 = (torch.Tensor(range(11)) - 5) / 5
    linear_multiclass_input = torch.cat((linear_input_0_1.unsqueeze(1), linear_input_minus1_1.unsqueeze(1)), dim=1)
    ece_losses = [BECELoss, BECEWithLogitsLoss, ExclusiveCrossEntropyLoss]
    inputs = [linear_input_0_1, linear_input_minus1_1, linear_multiclass_input]
    expected_outputs = [
        [0.0000, 0.0167, 0.0380, 0.0632, 0.0932, 0.1293, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000],
        [0.0549, 0.0660, 0.0789, 0.0937, 0.1105, 0.1293, 0.1504, 0.1735, 0.0000, 0.0000, 0.0000],
        [0.0000, 0.0000, 0.0000, 0.0000, 0.0000, 0.0860, 0.0937, 0.1018, 0.1105, 0.1196, 0.1293],
    ]
    for input_sample, expected_output, ece_loss in zip(inputs, expected_outputs, ece_losses):
        input_sample.requires_grad = True
        expected_output = torch.tensor(expected_output)

        loss = ece_loss()
        loss.set_epoch(300)
        loss.sample_from_background = _identity_mapping_single_input_torch_zeros  # all background (fully un-annotated)

        output = loss(input_sample).detach()
        if not torch.allclose(expected_output, output, atol=1e-04):
            print(expected_output, output)
        assert torch.allclose(expected_output, output, atol=1e-04)


def test_beceloss():
    """Basic functionality for BECE"""
    loss = BECELoss()
    batch_size = 32
    epoch = 100
    input_sample = torch.rand(batch_size, requires_grad=True)
    output = loss(input_sample, epoch=epoch)
    assert output.shape == input_sample.shape
    assert not output.isnan().any()


def test_becewithlogitsloss():
    """Basic functionality for BECEWithLogits"""
    loss = BECEWithLogitsLoss()
    batch_size = 32
    epoch = 100
    input_sample = torch.rand(batch_size, requires_grad=True)
    output = loss(input_sample, epoch=epoch)
    assert output.shape == input_sample.shape
    assert not output.isnan().any()


def test_exclusivecrossentropyloss():
    """Basic functionality for ECE"""
    loss = ExclusiveCrossEntropyLoss()
    batch_size = 32
    epoch = 100
    num_classes = 10
    input_sample = torch.randn(batch_size, num_classes, requires_grad=True)
    target = torch.empty(batch_size, dtype=torch.long).random_(num_classes)
    output = loss(input_sample, target, epoch)
    assert output.shape == target.shape
    assert not output.isnan().any()

    output = loss(input_sample, epoch=epoch)
    assert not output.isnan().any()


def test_example_beceloss():
    """The docstring example of BECE"""
    sigmoid = torch.nn.Sigmoid()
    loss = BECELoss()
    input_sample = torch.randn(3, requires_grad=True)
    target = torch.empty(3).random_(2)
    output = loss(sigmoid(input_sample), target)
    assert output.shape == target.shape
    assert not output.isnan().any()


def test_example_becelosswithlogits():
    """The docstring example of BECEWithLogits"""
    loss = BECEWithLogitsLoss()
    batch_size = torch.randint(32, (1, 1)).squeeze()
    input_sample = torch.rand(batch_size, requires_grad=True)
    output = loss(input_sample)
    assert output.shape == input_sample.shape
    assert not output.isnan().any()
    output = loss(input_sample).mean()


def test_example_exclusivecrossentropy():
    """The docstring example of ECE"""
    loss = ExclusiveCrossEntropyLoss()
    epoch = torch.randint(300, (1, 1)).squeeze()
    batch_size = torch.randint(32, (1, 1)).squeeze()
    num_classes = torch.randint(100, (1, 1)).squeeze() + 1
    input_sample = torch.randn(batch_size, num_classes, requires_grad=True)
    target = torch.empty(batch_size, dtype=torch.long).random_(num_classes)
    output = loss(input_sample, target, epoch)
    assert output.shape == target.shape
    assert not output.isnan().any()
    output = torch.mean(output)


def test_device():  # pylint: disable=R0914
    """The device inputs are handled corrects"""
    batch_size = 32
    num_classes = 10

    losses = [BECELoss, BECEWithLogitsLoss, ExclusiveCrossEntropyLoss]
    input_sizes = [(batch_size,), (batch_size,), (batch_size, num_classes)]
    target_sizes = [(batch_size,), (batch_size,), (batch_size,)]
    target_dtypes = [torch.float, torch.float, torch.long]
    devices = ["cpu", "gpu"]
    for input_size, target_size, loss_, target_dtype in zip(input_sizes, target_sizes, losses, target_dtypes):
        for device in devices:
            loss = loss_()
            predictions = torch.rand(input_size, requires_grad=True)
            target = torch.zeros(target_size, dtype=target_dtype)
            if device == "cpu":
                predictions = predictions.cpu()
                target = target.cpu()
            elif device == "gpu":
                if not torch.cuda.is_available():
                    print("No CUDA available, skip GPU testing.")
                    continue
                predictions = predictions.cuda()
                target = target.cuda()
            else:
                raise NotImplementedError

            output = loss(predictions)
            assert not output.isnan().any()
            output = loss(predictions, target)
            assert not output.isnan().any()



#!/usr/bin/env python
# coding=utf-8

"""The setup script."""

import ast

from setuptools import find_packages, setup  # type: ignore

with open("ece_loss/__init__.py", "r", encoding="utf-8") as f:
    for line in f:
        if line.startswith("__version__"):
            version = ast.parse(line).body[0].value.s  # type: ignore
            break


# Get the long description from the README file
with open("README.rst", "r", encoding="utf-8") as f:
    LONG_DESCRIPTION = f.read()


setup(
    author="Jonas Teuwen",
    author_email="j.teuwen@nki.nl",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/x-rst",
    python_requires=">=3.8",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    install_requires=["torch~=1.9.0", "setuptools~=58.0.4"],
    extras_require={
        "dev": ["numpy~=1.21.2", "pytest", "pylint", "black", "isort", "tox"],
        "example": ["pillow~=8.4.0", "matplotlib~=3.4.3"],
    },
    license="Apache Software License 2.0",
    include_package_data=True,
    keywords="ece_loss",
    name="ece_loss",
    packages=find_packages(include=["ece_loss", "ece_loss.*"]),
    url="https://github.com/NKI-AI/ece_loss",
    version=version,
    zip_safe=False,
)
